@@ -5,7 +5,7 @@
 %%% Created :  8 Aug 2009 by Evgeniy Khramtsov <ekhramtsov@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2020 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2021 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -50,45 +50,38 @@
 -export([rand_uniform/0, rand_uniform/1, rand_uniform/2, unmap_v4_addr/1]).
 
 -include("stun.hrl").
+-include("stun_logger.hrl").
 
 -define(MAX_BUF_SIZE, 64*1024). %% 64kb
 -define(TIMEOUT, 60000). %% 1 minute
 -define(NONCE_LIFETIME, 60*1000*1000). %% 1 minute (in usec)
 -define(SERVER_NAME, <<"P1 STUN library">>).
-%% RFC 6156, 9.1: "a TURN relay MUST NOT accept Teredo or 6to4 addresses".
--define(BLACKLIST, [{{8193, 0, 0, 0, 0, 0, 0, 0}, 32},   % 2001::/32 (Teredo).
-		    {{8194, 0, 0, 0, 0, 0, 0, 0}, 16}]). % 2002::/16 (6to4).
-
-%%-define(debug, true).
--ifdef(debug).
--define(dbg(Str, Args), error_logger:info_msg(Str, Args)).
--else.
--define(dbg(Str, Args), ok).
--endif.
 
 -type addr() :: {inet:ip_address(), inet:port_number()}.
 
 -record(state,
-	{sock                        :: inet:socket() | fast_tls:tls_socket(),
+	{sock                        :: inet:socket() | fast_tls:tls_socket() | undefined,
 	 sock_mod = gen_tcp          :: gen_udp | gen_tcp | fast_tls,
-	 certfile                    :: iodata(),
+	 certfile                    :: iodata() | undefined,
 	 peer = {{0,0,0,0}, 0}       :: addr(),
-	 tref = make_ref()           :: reference(),
+	 tref                        :: reference() | undefined,
 	 use_turn = false            :: boolean(),
 	 relay_ipv4_ip = {127,0,0,1} :: inet:ip4_address(),
-	 relay_ipv6_ip               :: inet:ip6_address(),
+	 relay_ipv6_ip               :: inet:ip6_address() | undefined,
 	 min_port = 49152            :: non_neg_integer(),
 	 max_port = 65535            :: non_neg_integer(),
 	 max_allocs = 10             :: non_neg_integer() | infinity,
 	 shaper = none               :: stun_shaper:shaper(),
 	 max_permissions = 10        :: non_neg_integer() | infinity,
-	 blacklist = ?BLACKLIST      :: turn:blacklist(),
+	 blacklist = []              :: turn:blacklist(),
 	 auth = user                 :: anonymous | user,
 	 nonces = treap:empty()      :: treap:treap(),
 	 realm = <<"">>              :: binary(),
-	 auth_fun                    :: function(),
+	 auth_fun                    :: function() | undefined,
+	 hook_fun                    :: function() | undefined,
 	 server_name = ?SERVER_NAME  :: binary(),
-	 buf = <<>>                  :: binary()}).
+	 buf = <<>>                  :: binary(),
+	 session_id                  :: binary() | undefined}).
 
 %%====================================================================
 %% API
@@ -115,9 +108,10 @@ udp_recv(Sock, Addr, Port, Data, State) ->
     NewState = prepare_state(State, Sock, {Addr, Port}, gen_udp),
     case stun_codec:decode(Data, datagram) of
  	{ok, Msg} ->
- 	    ?dbg("got: ~s", [stun_codec:pp(Msg)]),
+ 	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
  	    process(NewState, Msg);
- 	_ ->
+ 	Err ->
+	    ?LOG_DEBUG("Cannot parse packet: ~p", [Err]),
 	    NewState
     end.
 
@@ -125,26 +119,31 @@ udp_recv(Sock, Addr, Port, Data, State) ->
 %% gen_fsm callbacks
 %%====================================================================
 init([Sock, Opts]) ->
+    process_flag(trap_exit, true),
     case inet:peername(Sock) of
 	{ok, Addr} ->
-	    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
-	    SockMod = get_sockmod(Opts),
-	    State = prepare_state(Opts, Sock, Addr, SockMod),
-	    CertFile = get_certfile(Opts),
-	    case maybe_starttls(Sock, SockMod, CertFile, Addr) of
-		{ok, NewSock} ->
-		    inet:setopts(Sock, [{active, once}]),
-		    {ok, session_established,
-		     State#state{tref = TRef, sock = NewSock}};
-		{error, Why} ->
-		    {stop, Why}
+	    case get_sockmod(Opts, Sock) of
+		{ok, SockMod} ->
+		    State = prepare_state(Opts, Sock, Addr, SockMod),
+		    CertFile = get_certfile(Opts),
+		    case maybe_starttls(Sock, SockMod, CertFile) of
+			{ok, NewSock} ->
+			    inet:setopts(Sock, [{active, once}]),
+			    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
+			    {ok, session_established,
+			     State#state{tref = TRef, sock = NewSock}};
+			{error, Reason} ->
+			    {stop, Reason}
+		    end;
+		{error, Reason} ->
+		    {stop, Reason}
 	    end;
 	Err ->
-	    Err
+	    {stop, Err}
     end.
 
 session_established(Event, State) ->
-    error_logger:error_msg("unexpected event in session_established: ~p", [Event]),
+    ?LOG_ERROR("Unexpected event in 'session_established': ~p", [Event]),
     {next_state, session_established, State}.
 
 handle_event(stop, _StateName, State) ->
@@ -161,26 +160,28 @@ handle_info({tcp, _Sock, TLSData}, StateName,
     case fast_tls:recv_data(NewState#state.sock, TLSData) of
 	{ok, Data} ->
 	    process_data(StateName, NewState, Data);
-	_Err ->
+	Err ->
+	    ?LOG_INFO("Connection failure: ~p", [Err]),
 	    {stop, normal, NewState}
     end;
 handle_info({tcp, _Sock, Data}, StateName, State) ->
     NewState = update_shaper(State, Data),
     process_data(StateName, NewState, Data);
 handle_info({tcp_closed, _Sock}, _StateName, State) ->
-    ?dbg("connection reset by peer", []),
+    ?LOG_INFO("Connection reset by peer"),
     {stop, normal, State};
 handle_info({tcp_error, _Sock, _Reason}, _StateName, State) ->
-    ?dbg("connection error: ~p", [_Reason]),
+    ?LOG_INFO("Connection error: ~p", [_Reason]),
     {stop, normal, State};
 handle_info({timeout, TRef, stop}, _StateName,
 	    #state{tref = TRef} = State) ->
+    ?LOG_INFO("Connection timed out"),
     {stop, normal, State};
 handle_info({timeout, _TRef, activate}, StateName, State) ->
     activate_socket(State),
     {next_state, StateName, State};
 handle_info(Info, StateName, State) ->
-    error_logger:error_msg("unexpected info: ~p", [Info]),
+    ?LOG_ERROR("Unexpected info in '~s': ~p", [StateName, Info]),
     {next_state, StateName, State}.
 
 terminate(_Reason, _StateName, State) ->
@@ -216,6 +217,8 @@ process(#state{auth = anonymous} = State,
 	      'REALM' = Realm,
 	      'NONCE' = Nonce} = Msg)
   when User /= undefined, Realm /= undefined, Nonce /= undefined ->
+    ?LOG_NOTICE("Rejecting request: Credentials provided for anonymous "
+		"service"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(401)},
@@ -226,6 +229,7 @@ process(#state{auth = user} = State,
 	      'REALM' = Realm,
 	      'NONCE' = Nonce} = Msg)
   when User /= undefined, Realm /= undefined, Nonce /= undefined ->
+    stun_logger:add_metadata(#{stun_user => User}),
     Resp = prepare_response(State, Msg),
     {HaveNonce, Nonces} = have_nonce(Nonce, State#state.nonces),
     case HaveNonce of
@@ -237,29 +241,22 @@ process(#state{auth = user} = State,
 			  'NONCE' = Nonce},
 	    case (State#state.auth_fun)(User, Realm) of
 		<<"">> ->
-		    error_logger:info_msg(
-		      "failed long-term STUN authentication "
-		      "for ~s@~s from ~s",
-		      [User, Realm, addr_to_str(State#state.peer)]),
+		    ?LOG_NOTICE("Failed long-term STUN/TURN authentication"),
 		    send(NewState, R);
 		Pass ->
-		    Key = {User, Realm, Pass},
-		    case stun_codec:check_integrity(Msg, Key) of
-			true ->
-			    error_logger:info_msg(
-			      "accepted long-term STUN authentication "
-			      "for ~s@~s from ~s",
-			      [User, Realm, addr_to_str(State#state.peer)]),
+		    case check_integrity(User, Realm, Msg, Pass) of
+			{true, Key} ->
+			    ?LOG_INFO("Accepting long-term STUN/TURN "
+				      "authentication"),
 			    process(NewState, Msg, Key);
 			false ->
-			    error_logger:info_msg(
-			      "failed long-term STUN authentication "
-			      "for ~s@~s from ~s",
-			      [User, Realm, addr_to_str(State#state.peer)]),
+			    ?LOG_NOTICE("Failed long-term STUN/TURN "
+					"authentication"),
 			    send(NewState, R)
 		    end
 	    end;
 	false ->
+	    ?LOG_NOTICE("Rejecting request: Nonexistent nonce"),
 	    {NewNonce, NewNonces} = make_nonce(State#state.peer, Nonces),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(438),
@@ -271,11 +268,13 @@ process(State, #stun{class = request,
 		     'USERNAME' = User,
 		     'REALM' = undefined,
 		     'NONCE' = undefined} = Msg) when User /= undefined ->
+    ?LOG_NOTICE("Rejecting request: Missing realm and nonce"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(401)},
     send(State, R);
 process(State, #stun{class = request} = Msg) ->
+    ?LOG_NOTICE("Rejecting malformed request"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(400)},
@@ -288,10 +287,13 @@ process(State, Msg) when is_record(Msg, turn) ->
 process(State, _Msg) ->
     State.
 
-process(State, #stun{class = request, unsupported = [_|_]} = Msg, Secret) ->
+process(State, #stun{class = request, unsupported = [_|_] = Unsupported} = Msg,
+	Secret) ->
+    ?LOG_DEBUG("Rejecting request with unknown attribute(s): ~p",
+	       [Unsupported]),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
-		  'UNKNOWN-ATTRIBUTES' = Msg#stun.unsupported,
+		  'UNKNOWN-ATTRIBUTES' = Unsupported,
 		  'ERROR-CODE' = stun_codec:error(420)},
     send(State, R, Secret);
 process(State, #stun{class = request,
@@ -300,13 +302,17 @@ process(State, #stun{class = request,
     AddrPort = unmap_v4_addr(State#state.peer),
     R = case stun_codec:version(Msg) of
 	    old ->
+		?LOG_DEBUG("Responding to 'classic' STUN request"),
 		Resp#stun{class = response, 'MAPPED-ADDRESS' = AddrPort};
 	    new ->
+		?LOG_DEBUG("Responding to STUN request"),
 		Resp#stun{class = response, 'XOR-MAPPED-ADDRESS' = AddrPort}
 	end,
+    run_hook(stun_query, State, Msg),
     send(State, R, Secret);
 process(#state{use_turn = false} = State,
 	#stun{class = request} = Msg, Secret) ->
+    ?LOG_NOTICE("Rejecting TURN request: TURN is disabled"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
     send(State, R, Secret);
@@ -334,7 +340,9 @@ process(State, #stun{class = request,
 		    {relay_ipv4_ip, State#state.relay_ipv4_ip},
 		    {relay_ipv6_ip, State#state.relay_ipv6_ip},
 		    {min_port, State#state.min_port},
-		    {max_port, State#state.max_port} |
+		    {max_port, State#state.max_port},
+		    {hook_fun, State#state.hook_fun},
+		    {session_id, State#state.session_id} |
 		    if SockMod /= gen_udp ->
 			    [{owner, self()}];
 		       true ->
@@ -346,16 +354,17 @@ process(State, #stun{class = request,
 		    turn:route(Pid, Msg),
 		    State;
 		{error, limit} ->
+		    ?LOG_NOTICE("Rejecting request: Allocation quota reached"),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(486)},
 		    send(State, R, Secret);
 		{error, stale} ->
+		    ?LOG_NOTICE("Rejecting request: Stale nonce"),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(438)},
 		    send(State, R);
 		Err ->
-		    error_logger:error_msg(
-		      "failed to start turn session: ~p", [Err]),
+		    ?LOG_ERROR("Cannot start TURN session: ~p", [Err]),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(500)},
 		    send(State, R, Secret)
@@ -371,17 +380,16 @@ process(State, #stun{class = request,
 		    method = ?STUN_METHOD_CHANNEL_BIND} = Msg, Secret) ->
     route_on_turn(State, Msg, Secret);
 process(State, #stun{class = request} = Msg, Secret) ->
+    ?LOG_NOTICE("Rejecting request: Method not allowed"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
-    send(State, R, Secret);
-process(State, _Msg, _Secret) ->
-    State.
+    send(State, R, Secret).
 
 process_data(NextStateName, #state{buf = Buf} = State, Data) ->
     NewBuf = <<Buf/binary, Data/binary>>,
     case stun_codec:decode(NewBuf, stream) of
 	{ok, Msg, Tail} ->
-	    ?dbg("got:~n~s", [stun_codec:pp(Msg)]),
+ 	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
 	    NewState = process(State, Msg),
 	    process_data(NextStateName, NewState#state{buf = <<>>}, Tail);
 	empty ->
@@ -390,7 +398,8 @@ process_data(NextStateName, #state{buf = Buf} = State, Data) ->
 	more when size(NewBuf) < ?MAX_BUF_SIZE ->
 	    NewState = State#state{buf = NewBuf},
 	    {next_state, NextStateName, NewState};
-	_ ->
+	{error, Reason} ->
+	    ?LOG_DEBUG("Cannot parse packet: ~p", [Reason]),
 	    {stop, normal, State}
     end.
 
@@ -420,22 +429,13 @@ send(State, Data) when is_binary(Data) ->
 send(State, Msg) ->
     send(State, Msg, undefined).
 
-send(State, Msg, {_JID, Pass}) ->
-    send(State, Msg, Pass);
 send(State, Msg, Pass) ->
-    ?dbg("send:~n~s", [stun_codec:pp(Msg)]),
-    case Msg of
-	#stun{class = indication} ->
-	    send(State, stun_codec:encode(Msg, undefined));
-	_ ->
-	    send(State, stun_codec:encode(Msg, Pass))
-    end.
+    ?LOG_DEBUG(#{verbatim => {"Sending:~n~s", [stun_codec:pp(Msg)]}}),
+    send(State, stun_codec:encode(Msg, Pass)).
 
 route_on_turn(State, Msg) ->
     route_on_turn(State, Msg, undefined).
 
-route_on_turn(State, Msg, {_JID, Pass}) ->
-    route_on_turn(State, Msg, Pass);
 route_on_turn(State, Msg, Pass) ->
     case turn_sm:find_allocation(State#state.peer) of
 	{ok, Pid} ->
@@ -444,6 +444,7 @@ route_on_turn(State, Msg, Pass) ->
 	_ ->
 	    case Msg of
 		#stun{class = request} ->
+		    ?LOG_NOTICE("Rejecting request: Allocation mismatch"),
 		    Resp = prepare_response(State, Msg),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(437)},
@@ -454,20 +455,25 @@ route_on_turn(State, Msg, Pass) ->
     end.
 
 prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
+    ID = case proplists:get_value(session_id, Opts) of
+	     ID0 when is_binary(ID0) ->
+		 ID0; % Stick to listener's session ID.
+	     undefined ->
+		 stun_logger:make_id()
+	 end,
+    stun_logger:set_metadata(stun, SockMod, ID, Peer),
     case proplists:get_bool(use_turn, Opts) of
 	true ->
 	    lists:foldl(
 	      fun({turn_ip, IP}, State) ->
 		      case prepare_addr(IP) of
 			  {ok, Addr} ->
-			      error_logger:error_msg(
-				"'turn_ip' is deprecated, specify "
-				"'turn_ipv4_address' and optionally "
-				"'turn_ipv6_address' instead", []),
+			      ?LOG_WARNING("'turn_ip' is deprecated, specify "
+					   "'turn_ipv4_address' and optionally "
+					   "'turn_ipv6_address' instead"),
 			      State#state{relay_ipv4_ip = Addr};
 			  {error, _} ->
-			      error_logger:error_msg(
-				"wrong 'turn_ip' value: ~p", [IP]),
+			      ?LOG_ERROR("Wrong 'turn_ip' value: ~p", [IP]),
 			      State
 		      end;
 		 ({turn_ipv4_address, IP}, State) ->
@@ -475,8 +481,8 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 			  {ok, Addr} ->
 			      State#state{relay_ipv4_ip = Addr};
 			  {error, _} ->
-			      error_logger:error_msg(
-				"wrong 'turn_ipv4_address' value: ~p", [IP]),
+			      ?LOG_ERROR("Wrong 'turn_ipv4_address' value: ~p",
+					 [IP]),
 			      State
 		      end;
 		 ({turn_ipv6_address, IP}, State) ->
@@ -484,83 +490,81 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 			  {ok, Addr} ->
 			      State#state{relay_ipv6_ip = Addr};
 			  {error, _} ->
-			      error_logger:error_msg(
-				"wrong 'turn_ipv6_address' value: ~p", [IP]),
+			      ?LOG_ERROR("Wrong 'turn_ipv6_address' value: ~p",
+					 [IP]),
 			      State
 		      end;
 		 ({turn_min_port, Min}, State)
 		    when is_integer(Min), Min > 1024, Min < 65536 ->
 		      State#state{min_port = Min};
 		 ({turn_min_port, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'turn_min_port' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'turn_min_port' value: ~p", [Wrong]),
 		      State;
 		 ({turn_max_port, Max}, State)
 		    when is_integer(Max), Max > 1024, Max < 65536 ->
 		      State#state{max_port = Max};
 		 ({turn_max_port, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'turn_max_port' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'turn_max_port' value: ~p", [Wrong]),
 		      State;
 		 ({turn_max_allocations, N}, State)
 		    when (is_integer(N) andalso N > 0) orelse is_atom(N) ->
 		      State#state{max_allocs = N};
 		 ({turn_max_allocations, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'turn_max_allocations' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'turn_max_allocations' value: ~p",
+				 [Wrong]),
 		      State;
 		 ({turn_max_permissions, N}, State)
 		    when (is_integer(N) andalso N > 0) orelse is_atom(N) ->
 		      State#state{max_permissions = N};
+		 ({turn_max_permissions, Wrong}, State) ->
+		      ?LOG_ERROR("Wrong 'turn_max_permissions' value: ~p",
+				 [Wrong]),
+		      State;
 		 ({turn_blacklist, B}, State) ->
 		      case lists:all(fun is_valid_subnet/1, B) of
 			  true ->
 			      State#state{blacklist = B};
 			  false ->
-			      error_logger:error_msg(
-				"wrong 'turn_blacklist' value: ~p", [B]),
+			      ?LOG_ERROR("Wrong 'turn_blacklist' value: ~p",
+					 [B]),
 			      State
 		      end;
-		 ({turn_max_permissions, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'turn_max_permissions' value: ~p", [Wrong]),
-		      State;
 		 ({shaper, S}, State)
 		    when S == none orelse (is_integer(S) andalso (S > 0)) ->
 		      State#state{shaper = stun_shaper:new(S)};
 		 ({shaper, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'shaper' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'shaper' value: ~p", [Wrong]),
 		      State;
 		 ({server_name, S}, State) ->
 		      try
 			  State#state{server_name = iolist_to_binary(S)}
 		      catch _:_ ->
-			      error_logger:error_msg(
-				"wrong 'server_name' value: ~p", [S]),
+			      ?LOG_ERROR("Wrong 'server_name' value: ~p", [S]),
 			      State
 		      end;
 		 ({auth_realm, R}, State) ->
 		      try
 			  State#state{realm = iolist_to_binary(R)}
 		      catch _:_ ->
-			      error_logger:error_msg(
-				"wrong 'auth_realm' value: ~p", [R]),
+			      ?LOG_ERROR("Wrong 'auth_realm' value: ~p", [R]),
 			      State
 		      end;
 		 ({auth_fun, F}, State) when is_function(F) ->
 		      State#state{auth_fun = F};
 		 ({auth_fun, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'auth_fun' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'auth_fun' value: ~p", [Wrong]),
+		      State;
+		 ({hook_fun, F}, State) when is_function(F) ->
+		      State#state{hook_fun = F};
+		 ({hook_fun, Wrong}, State) ->
+		      ?LOG_ERROR("Wrong 'hook_fun' value: ~p", [Wrong]),
 		      State;
 		 ({auth_type, anonymous}, State) ->
 		      State#state{auth = anonymous};
 		 ({auth_type, user}, State) ->
 		      State#state{auth = user};
 		 ({auth_type, Wrong}, State) ->
-		      error_logger:error_msg(
-			"wrong 'auth_type' value: ~p", [Wrong]),
+		      ?LOG_ERROR("Wrong 'auth_type' value: ~p", [Wrong]),
 		      State;
 		 ({use_turn, _}, State) -> State;
 		 (use_turn, State) -> State;
@@ -570,19 +574,23 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 		 ({certfile, _}, State) -> State;
 		 ({tls, _}, State) -> State;
 		 (tls, State) -> State;
+		 ({session_id, _}, State) -> State;
 		 (Opt, State) ->
-		      error_logger:error_msg(
-			"ignoring unknown option ~p", [Opt]),
+		      ?LOG_ERROR("Ignoring unknown option '~p'", [Opt]),
 		      State
 	      end,
-	      #state{peer = Peer, sock = Sock,
+	      #state{session_id = ID, peer = Peer, sock = Sock,
 		     sock_mod = SockMod, use_turn = true},
 	      Opts);
 	_ ->
-	    #state{sock = Sock, sock_mod = SockMod, peer = Peer}
+	    #state{session_id = ID, sock = Sock, sock_mod = SockMod,
+		   peer = Peer, hook_fun = proplists:get_value(hook_fun, Opts),
+		   auth = anonymous}
     end;
-prepare_state(State, _Sock, Peer, _SockMod) ->
-    State#state{peer = Peer}.
+prepare_state(State, _Sock, Peer, SockMod) ->
+    ID = stun_logger:make_id(),
+    stun_logger:set_metadata(stun, SockMod, ID, Peer),
+    State#state{session_id = ID, peer = Peer}.
 
 prepare_addr(IPBin) when is_binary(IPBin) ->
     prepare_addr(binary_to_list(IPBin));
@@ -651,18 +659,24 @@ have_nonce(Nonce, Nonces) ->
 	    {false, NewNonces}
     end.
 
+check_integrity(User, Realm, Msg, Pass) when is_binary(Pass) ->
+    check_integrity(User, Realm, Msg, [Pass]);
+check_integrity(_User, _Realm, _Msg, []) ->
+    false;
+check_integrity(User, Realm, Msg, [Pass | T]) ->
+    Key = {User, Realm, Pass},
+    case stun_codec:check_integrity(Msg, Key) of
+	true ->
+	    {true, Key};
+	false ->
+	    check_integrity(User, Realm, Msg, T)
+    end.
+
 unmap_v4_addr({{0, 0, 0, 0, 0, 16#FFFF, D7, D8}, Port}) ->
     {{D7 bsr 8, D7 band 255, D8 bsr 8, D8 band 255}, Port};
 unmap_v4_addr(AddrPort) ->
     AddrPort.
 
-addr_to_str({{_, _, _, _, _, _, _, _} = Addr, Port}) ->
-    [$[, inet_parse:ntoa(Addr), $], $:, integer_to_list(Port)];
-addr_to_str({{_, _, _, _} = Addr, Port}) ->
-    [inet_parse:ntoa(Addr), $:, integer_to_list(Port)];
-addr_to_str(Addr) ->
-    inet_parse:ntoa(Addr).
-
 is_valid_subnet({{IP1, IP2, IP3, IP4}, Mask}) ->
     (IP1 >= 0) and (IP1 =< 255) and
     (IP2 >= 0) and (IP2 =< 255) and
@@ -682,12 +696,21 @@ is_valid_subnet({{IP1, IP2, IP3, IP4, IP5, IP6, IP7, IP8}, Mask}) ->
 is_valid_subnet(_) ->
     false.
 
-get_sockmod(Opts) ->
-    case proplists:get_bool(tls, Opts) of
+get_sockmod(Opts, Sock) ->
+    case proplists:get_value(tls, Opts, false) of
 	true ->
-	    fast_tls;
+	    {ok, fast_tls};
 	false ->
-	    gen_tcp
+	    {ok, gen_tcp};
+	optional ->
+	    case is_tls_handshake(Sock) of
+		true ->
+		    {ok, fast_tls};
+		false ->
+		    {ok, gen_tcp};
+		{error, _Reason} = Err ->
+		    Err
+	    end
     end.
 
 get_certfile(Opts) ->
@@ -698,14 +721,32 @@ get_certfile(Opts) ->
 	    undefined
     end.
 
-maybe_starttls(_Sock, fast_tls, undefined, {IP, Port}) ->
-    error_logger:error_msg("failed to start TLS connection for ~s:~p: "
-			   "option 'certfile' is not set",
-			   [inet_parse:ntoa(IP), Port]),
+-ifdef(USE_OLD_INET_BACKEND).
+-dialyzer({[no_match], [get_sockmod/2]}).
+is_tls_handshake(_Sock) ->
+    ?LOG_ERROR("Multiplexing TCP and TLS requires a newer Erlang/OTP version"),
+    {error, eprotonosupport}.
+-else.
+is_tls_handshake({_, _, {_, Socket}}) ->
+    case socket:recvfrom(Socket, 10, [peek], ?TIMEOUT) of
+	{ok, {_, <<22, 3, _:4/binary, 0, _:2/binary, 3>>}} ->
+	    ?LOG_DEBUG("Determined transport protocol: TLS"),
+	    true;
+	{ok, {_, _}} ->
+	    ?LOG_DEBUG("Determined transport protocol: TCP"),
+	    false;
+	{error, Reason} = Err ->
+	    ?LOG_INFO("Cannot determine transport protocol: ~s", [Reason]),
+	    Err
+    end.
+-endif.
+
+maybe_starttls(_Sock, fast_tls, undefined) ->
+    ?LOG_ERROR("Cannot start TLS connection: option 'certfile' is not set"),
     {error, eprotonosupport};
-maybe_starttls(Sock, fast_tls, CertFile, _PeerAddr) ->
-    fast_tls:tcp_to_tls(Sock, [{certfile, CertFile}]);
-maybe_starttls(Sock, gen_tcp, _CertFile, _PeerAddr) ->
+maybe_starttls(Sock, fast_tls, CertFile) ->
+    fast_tls:tcp_to_tls(Sock, [verify_none, {certfile, CertFile}]);
+maybe_starttls(Sock, gen_tcp, _CertFile) ->
     {ok, Sock}.
 
 prepare_response(State, Msg) ->
@@ -714,6 +755,28 @@ prepare_response(State, Msg) ->
 	  trid = Msg#stun.trid,
 	  'SOFTWARE' = State#state.server_name}.
 
+run_hook(HookName,
+	 #state{session_id = ID,
+		peer = Client,
+		sock_mod = SockMod,
+		hook_fun = HookFun},
+	 #stun{'USERNAME' = User,
+	       'REALM' = Realm} = Msg)
+  when is_function(HookFun) ->
+    Info = #{id => ID,
+	     user => User,
+	     realm => Realm,
+	     client => Client,
+	     transport => stun_logger:encode_transport(SockMod),
+	     version => stun_codec:version(Msg)},
+    ?LOG_DEBUG("Running '~s' hook", [HookName]),
+    try HookFun(HookName, Info)
+    catch _:Err -> ?LOG_ERROR("Hook '~s' failed: ~p", [HookName, Err])
+    end;
+run_hook(HookName, _State, _Msg) ->
+    ?LOG_DEBUG("No callback function specified for '~s' hook", [HookName]),
+    ok.
+
 -define(THRESHOLD, 16#10000000000000000).
 
 -ifdef(RAND_UNIFORM).
